import typing
from datetime import datetime


class Song:
    def __init__(self, **data):
        if not data:
            raise LookupError("No data was given.")

        self.data = data

    # variables, adding them as properties so they can't be changed
    @property
    def title(self) -> str:
        return self.data["title"]

    @property
    def name(self) -> str:
        return self.title

    @property
    def artists(self) -> typing.List[str]:
        return self.data["artists"]

    @property
    def versions(self) -> typing.List[str]:
        return self.data["versions"]

    @property
    def release_date(self) -> datetime:
        ret = self.data["release_date"]

        return datetime.strptime(ret, "%d %b %Y")

    @property
    def artwork(self) -> str:
        return self.data["artwork"]

    def __repr__(self):
        return f'<{type(self).__name__} title="{self.title}">'


class SearchResults:
    def __init__(self, pages: int, items: typing.List[dict]):
        self.pages = pages
        self.items: typing.List[Song] = []
        for item in items:
            self.items.append(Song(**item))

    def __repr__(self):
        return f"<{type(self).__name__} items={len(self.items)} pages={self.pages}>"



import asyncio
import aiohttp
import typing
from urllib.parse import quote_plus as urlencode

from .song import SearchResults, Song
from .parser import parse_search_results, parse_song, get_title


class Client:
    def __init__(self, loop=None, session=None, *, close_session=True):
        self.loop = loop or asyncio.get_event_loop()
        self.session = session or aiohttp.ClientSession(loop=self.loop)
        self.created_session = session is None and close_session

    async def _close_session(self, do=True):
        if self.created_session and do:
            await self.session.close()
            self.session = None

    async def get_song(self, id: str, close=True):
        if (self.created_session and close) and self.session == None:
            self.session = aiohttp.ClientSession(loop=self.loop)

        title_req = await self.session.get(f"https://ncs.io/{id}")
        title_req.raise_for_status()

        title_content = await title_req.content.read()
        title = await self.loop.run_in_executor(None, lambda: get_title(title_content))

        search = await self.search(
            title["title"], artists=title["artists"], get=False, close=False
        )
        info: Song = None
        for item in search.items:
            await asyncio.sleep(0)
            if item.title == title["title"]:
                info = item
                break
        
        if info == None:
            await self.close_session(close)
            return

        req = await self.session.get(
            f"https://ncs.lnk.to/{id}/widget?view=clickthrough"
        )
        req.raise_for_status()

        content = await req.content.read()
        data = await self.loop.run_in_executor(
            None, lambda: parse_song(content, searched=info.data)
        )

        await self.close_session(close)
        return Song(**data)

    async def search(
        self,
        query: str = None,
        *,
        genre: str = None,
        mood: str = None,
        artists: typing.List[str] = [],
        artist: str = None,
        version=-1,
        page: int = 1,
        get: bool = True,
        close=True,
    ) -> SearchResults:
        versions = {-1: "", 0: "regular-instrumental", 1: "regular", 2: "instrumental"}
        if not query and not genre and not mood and not artists and not artist:
            raise ValueError("Please give at least query.")

        genre = genre or ""
        mood = mood or ""
        artist = artist or ""

        if not isinstance(artists, list):
            raise ValueError("Artists must be a list.")

        if not isinstance(artist, str):
            raise ValueError("Artist must be one string.")

        if not isinstance(page, int):
            raise ValueError("Page must be an integer (number).")

        if not version in versions.keys() and not version.lower() in versions.values():
            vs = []
            for k, v in versions.items():
                vs.append(f'"{k}" or "{v}"')

            vs = ";\n".join(vs)
            raise ValueError(f"Version must be one of those:\n{vs}")
        else:
            version = versions[version] if isinstance(version, int) else version.lower()

        if artists and artist:
            artists = set(artists)
            artists.add(artist)
            artists = list(artists)

        if (self.created_session and close) and self.session == None:
            self.session = aiohttp.ClientSession(loop=self.loop)

        page = await self.session.get(
            "https://ncs.io/music-search?"
            f"q={urlencode(query)}&"
            f"genre={urlencode(genre)}&"
            f"mood={urlencode(mood)}&"
            f"version={urlencode(version)}&"
            f"page={page}"
        )
        page.raise_for_status()
        data = await page.content.read()
        data = await self.loop.run_in_executor(
            None, lambda: parse_search_results(data, wanted_artists=artists)
        )

        items = []
        if get:
            for item in data["items"]:
                items.append((await self.get_song(item["id"], close=False)).data)
        
            data["items"] = items

        
        await self.close_session(close)
        return SearchResults(**data)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ncspy",
    version="0.2.3",
    author="Vincy.zsh",
    author_email="Vincysuper07@gmail.com",
    description='A NoCopyrightSounds "API" wrapper written with asyncio.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Vincydotzsh/pyncs",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=["aiohttp", "beautifulsoup4", "slimit"],
    packages=["ncs"],
    python_requires=">=3.7",
)



import re

import slimit
from slimit import ast
from slimit.parser import Parser
from slimit.visitors import nodevisitor

try:
    from BeautifulSoup import BeautifulSoup
except ImportError:
    from bs4 import BeautifulSoup

# suppress useless ply warnings
slimit.lexer.ply.lex.PlyLogger = slimit.parser.ply.yacc.PlyLogger = type(
    "_NullLogger",
    (slimit.lexer.ply.lex.NullLogger,),
    dict(__init__=lambda s, *_, **__: (None, s.super().__init__())[0]),
)
parser = Parser()


def get_title(html):
    parsed = BeautifulSoup(html, features="html.parser")
    player = parsed.body.find("div", attrs={"class": "fake-table-cell"})
    artists = player.h2.span.text.strip().split(", ")
    player.h2.span.replace_with("")
    name = player.h2.text.strip()

    return {"artists": artists, "title": name}


def parse_song(html, searched=None) -> dict:
    parsed = BeautifulSoup(html, features="html.parser")
    marso = parsed.body.find("script", attrs={"id": "linkfire-widget-data"})
    tree = parser.parse(marso.text)
    fields = {
        getattr(node.left, "value", ""): getattr(node.right, "value", "")
        for node in nodevisitor.visit(tree)
        if isinstance(node, ast.Assign)
    }

    artwork = fields["artwork"].strip('" ').replace("\\", "")
    pagetitle = parsed.head.title.text
    comp = re.compile("(\((ft|Ft|feat|Feat)(|\.) .*\))")

    fts = comp.findall(pagetitle)
    feats = set()
    artists = set()
    for ft in fts:
        if isinstance(ft, tuple):
            ft = ft[0]

        ft = ft.strip("( )").replace("ft. ", "", 1).split(" & ")

        for ft in ft:
            feats.add(ft)

    datas = pagetitle.split(" - ")
    artists_ = datas[0].strip("( )").replace("ft. ", "", 1).split(" & ")
    for artist in artists_:
        if ", " in artist:
            for a in artist.split(", "):
                artists.add(a)
        else:
            artists.add(artist)

    feats = list(feats)
    artists = list(artists) + feats
    for ft in fts:
        if isinstance(ft, tuple):
            ft = ft[0]

        datas[1] = datas[1].replace(ft, "").strip()

    if len(datas) == 1:
        title = datas[0]
        artists = []
    else:
        title = datas[1]

    ret = {"title": title, "artists": artists, "artwork": artwork}
    if not searched:
        return ret

    ars_ = list(set(searched["artists"] + ret["artists"]))
    lows = []
    ars = []
    for ar in ars_:
        if not ar.lower() in lows:
            lows.append(ar.lower())
            ars.append(ar)

    searched.update(ret)
    searched["artists"] = ars
    return searched


def parse_search_results(html, wanted_artists=[]) -> dict:
    data = {"items": []}
    parsed = BeautifulSoup(html, features="html.parser")
    tbody = parsed.body.find("table", attrs={"class": "tablesorter"})
    if tbody == None:
        data["pages"] = 0
        return data

    tbody = tbody.tbody
    results = tbody.find_all("tr")
    for res in results:

        datas = res.find_all("td")[3]
        id = datas.a["href"].strip("/")
        title = datas.a.p.text
        artists = datas.span.text.split(", ")
        artists_ = datas.span.text.lower().split(", ")
        versions = res.find_all("td")[6].text.lower().split(", ")
        date = res.find_all("td")[5].text
        if not all(item.lower() in artists_ for item in wanted_artists):
            continue

        data["items"].append(
            {
                "id": id,
                "title": title,
                "artists": artists,
                "versions": versions,
                "release_date": date,
            }
        )

    pages = (parsed.body.find("ul", attrs={"class": "pagination"})) or 1
    if pages != 1:
        pages = pages.find_all("li", attrs={"class": "page-item"})[-2].text

    data["pages"] = pages
    return data
